@@ -15,20 +15,24 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
 """Test for utils.py module."""
-
+import json
+import pathlib
+import re
 from unittest.mock import Mock
 
-import matplotlib.pyplot as plt
 import numpy as np
 import pytest
 import requests
+import responses
+from PIL import Image
 
+from atldld.config import user_cache_dir
 from atldld.utils import (
-    CACHE_FOLDER,
     CommonQueries,
     get_2d,
     get_2d_bulk,
     get_3d,
+    get_corners_in_ref_space,
     get_experiment_list_from_gene,
     get_image,
     pir_to_xy_API_single,
@@ -39,46 +43,142 @@ EXISTING_IMAGE_IDS = [69750516, 101349546]
 EXISTING_DATASET_IDS = [479, 1357]  # [Gad, Gfap]
 
 
-class TestUtils:
-    def test_cache_folder_default(self):
-        """Test whether default cache path ends with '/'."""
-        assert CACHE_FOLDER.endswith("/")
-
+class TestGetImage:
     @pytest.mark.parametrize("image_id", EXISTING_IMAGE_IDS)  # too slow to try more
     def test_get_image_online(self, image_id, tmpdir, mocker):
-        """A test for the get_image function"""
-
-        mocker.patch("os.path.exists", return_value=True)
-        mocker.patch(
-            "matplotlib.pyplot.imread",
-            return_value=np.zeros((100, 200), dtype=np.uint8),
-        )
+        mocker.patch("pathlib.Path.exists", return_value=True)
+        fake_img = Image.fromarray(np.zeros((100, 200), dtype=np.uint8))
 
+        # Test getting the normal image
+        fake_img.save(pathlib.Path(tmpdir) / f"{image_id}-0.jpg")
         img = get_image(image_id, tmpdir)
         assert isinstance(img, np.ndarray)
         assert img.dtype == np.uint8
 
-        # Retrieve expression of the specified image
+        # Test getting the expression image
+        fake_img.save(pathlib.Path(tmpdir) / f"{image_id}-0-expression.jpg")
         img = get_image(image_id, tmpdir, expression=True)
         assert isinstance(img, np.ndarray)
         assert img.dtype == np.uint8
 
     def test_get_image_offline(self, tmpdir, img):
         """Test whether offline loading works"""
-
         image_id_fake = 123456
 
-        tmpdir_str = str(tmpdir)
-        tmpdir_str += "" if tmpdir_str.endswith("/") else "/"
+        img_path = pathlib.Path(tmpdir) / f"{image_id_fake}-0.jpg"
+        Image.fromarray(img, mode="L").save(img_path)
 
-        path_str = "{}{}.jpg".format(tmpdir_str, image_id_fake)
-
-        plt.imsave(path_str, img)
-
-        img = get_image(image_id_fake, tmpdir_str)
+        img = get_image(image_id_fake, str(tmpdir))
         assert isinstance(img, np.ndarray)
         assert img.dtype == np.uint8
 
+    def test_decompression_bomb_warning_suppressed(self, tmpdir):
+        image_id = 123
+
+        # Prepare a cached image. The decompression bomb warning is triggered
+        # at about 90M pixels, so we prepare an image with 100M pixels.
+        img = Image.fromarray(np.zeros((10_000, 10_000), dtype=np.uint8))
+        img.save(pathlib.Path(tmpdir) / f"{image_id}-0.jpg")
+
+        with pytest.warns(None) as records:
+            get_image(image_id, folder=str(tmpdir))
+
+        assert len(records) == 0
+
+    def test_decompression_bomb_error_triggered(self, tmpdir):
+        image_id = 123
+
+        # Prepare a cached image. The decompression bomb error is triggered
+        # at about 180M pixels, so we prepare an image with 200M pixels.
+        img = Image.fromarray(np.zeros((20_000, 10_000), dtype=np.uint8))
+        img.save(pathlib.Path(tmpdir) / f"{image_id}-0.jpg")
+
+        with pytest.raises(Image.DecompressionBombError):
+            get_image(image_id, folder=str(tmpdir))
+
+
+def test_get_corners_in_ref_space(mocker):
+    image_id = 123
+    width = 200
+    height = 100
+
+    def xy_to_pir(x, y, _image_id):
+        return x * 10, y * 10, 53
+
+    mocker.patch("atldld.utils.xy_to_pir_API_single", xy_to_pir)
+
+    corners = get_corners_in_ref_space(image_id, width, height)
+    assert isinstance(corners, np.ndarray)
+    assert np.all(
+        corners
+        == [
+            (0, 0, 53),
+            (width * 10, 0, 53),
+            (width * 10, height * 10, 53),
+            (0, height * 10, 53),
+        ]
+    )
+
+
+class TestXyToPirApiSingle:
+    @pytest.mark.internet
+    @pytest.mark.parametrize("image_id", EXISTING_IMAGE_IDS)
+    def test_fetching_online_no_mocking(self, image_id):
+        x, y = np.random.randint(0, 10000, 2)
+        p, i, r = xy_to_pir_API_single(x, y, image_id=image_id)
+        assert np.isfinite(p)
+        assert np.isfinite(i)
+        assert np.isfinite(r)
+
+    @responses.activate
+    def test_fetching_online_and_storing_in_cache(self):
+        # Mock the online response
+        image_id = 123
+        x, y = 10, 40
+        pir = (-983.5501122512056, 335.1538167934086, 4117.19706627407)
+        rv = {
+            "success": True,
+            "id": 0,
+            "start_row": 0,
+            "num_rows": 1,
+            "total_rows": 1,
+            "msg": {"image_to_reference": {"x": pir[0], "y": pir[1], "z": pir[2]}},
+        }
+        responses.add(responses.GET, re.compile(r".*/image_to_reference/.*"), json=rv)
+
+        # Check the returned value
+        pir_got = xy_to_pir_API_single(x, y, image_id=image_id)
+        assert pir_got == pir
+
+        # Check that the cache is updated
+        cache_file = user_cache_dir() / "image-to-reference" / f"{image_id}.json"
+        assert cache_file.exists()
+        with cache_file.open() as fp:
+            cache_json = json.load(fp)
+        assert cache_json.get(f"x={x}&y={y}") == list(pir)
+
+    @responses.activate
+    def test_reading_from_cache(self):
+        # Because we activated responses but didn't add any response values any
+        # request will raise an error. This way we automatically test that no
+        # requests are sent and the values are obtained offline from the cache
+        image_id = 123
+        x, y = 5, 10
+        pir = 20.5, 13.7, 8.8
+
+        # Prepare the cache file
+        cache_file = user_cache_dir() / "image-to-reference" / f"{image_id}.json"
+        cache_file.parent.mkdir()
+        cache_json = {f"x={x}&y={y}": pir}
+        with cache_file.open("w") as fp:
+            json.dump(cache_json, fp)
+
+        # Run the test
+        pir_got = xy_to_pir_API_single(x, y, image_id=image_id)
+        assert pir_got == pir
+
+
+class TestUtils:
     @pytest.mark.internet
     def test_get_experiment_list(self):
         """Test that experiment list is good."""
@@ -451,45 +551,6 @@ class TestUtils:
         assert np.isfinite(closest_section_image_id)
         assert isinstance(closest_section_image_id, int)
 
-    @pytest.mark.parametrize("image_id", EXISTING_IMAGE_IDS[:1])
-    @pytest.mark.parametrize("x", [10])
-    @pytest.mark.parametrize("y", [40])
-    def test_xy_to_pir_API_single(self, image_id, x, y, mocker):
-        """Test that xy to pir API works."""
-
-        # patching
-        fake_response = Mock(spec=requests.Request)
-        fake_response.ok = True
-
-        rv = {
-            "success": True,
-            "id": 0,
-            "start_row": 0,
-            "num_rows": 0,
-            "total_rows": 0,
-            "msg": {
-                "image_to_reference": {
-                    "x": -983.5501122512056,
-                    "y": 335.1538167934086,
-                    "z": 4117.19706627407,
-                }
-            },
-        }
-
-        fake_response.json = Mock(return_value=rv)
-        mocker.patch("requests.get", return_value=fake_response)
-
-        p, i, r = xy_to_pir_API_single(x, y, image_id=image_id)
-
-        # Mock calls
-        assert (
-            fake_response.json.call_count == 1
-        )  # assert_called not available in python 3.5 :(
-
-        assert np.isfinite(p)
-        assert np.isfinite(i)
-        assert np.isfinite(r)
-
     @pytest.mark.internet
     @pytest.mark.parametrize("dataset_id", EXISTING_DATASET_IDS)
     @pytest.mark.parametrize("p", [100, 100.1])
@@ -523,7 +584,7 @@ class TestUtils:
         assert closest_section_image_id == closest_section_image_id_s
 
 
-# @pytest.mark.internet
+@pytest.mark.internet
 class TestCommonQueries:
     """A set of tests focused on some basic queries in the CommonQueries class."""
 
